#!/usr/bin/env python3

# many thanks to Kareeeeem for the seizure-inducing glitch effects
#  github.com/Kareeeeem

import io
import copy
import random

from itertools import tee
from PIL import Image


class JpegError(Exception):
    pass


def pairwise(iterable):
    """Awesome function from the itertools cookbook
    https://docs.python.org/2/library/itertools.html
    s -> (s0,s1), (s1,s2), (s2, s3), ...
    """
    a, b = tee(iterable)
    next(b, None)
    return zip(a, b)


class Jpeg(object):

    def __init__(self, image_bytes, amount=random.randint(0,99), seed=random.randint(0,99), iterations=random.randint(0,115)):
        self.bytes = image_bytes
        self.new_bytes = None
        self.header_length = self.get_header_length()

        self.amount     = amount
        self.seed       = seed
        self.iterations = iterations


    def get_header_length(self):
        """Get the length of the header by searching sequential 0xFF 0xDA
        values. These values mark the end of a Jpeg header. We add two to give
        us a little leeway. We don't want to mess with the header.
        """

        for i, pair in enumerate(pairwise(self.bytes)):
            if pair[0] == 255 and pair[1] == 218:
                result = i + 2
                return result

        raise JpegError('Not a valid JPEG')


    def glitch_bytes(self):
        """Glitch the image bytes, after the header based on the parameters.
        'Amount' is the hex value that will be written into the file. 'Seed'
        tweaks the index where the value will be inserted, rather than just a
        simple division by iterations. 'Iterations' should be self explanatory
        """

        amount = self.amount / 100
        seed = self.seed / 100
        iterations = self.iterations

        # work with a copy of the original bytes. We might need the original
        # bytes around if we glitch it so much we break the file.
        new_bytes = copy.copy(self.bytes)

        for i in (range(iterations)):
            max_index = len(self.bytes) - self.header_length - 4

            # The following operations determine where we'll overwrite a value
            # Illustrate by example

            # 36 = (600 / 50) * 3
            px_min = int((max_index / iterations) * i)

            # 48 = (600 / 50) * 3 + 1
            px_max = int((max_index / iterations) * (i + 1))

            # 12 = 48 - 36
            delta = (px_max - px_min)  # * 0.8

            # 36 + (12 * 0.8)
            px_i = int(px_min + (delta * seed))

            # If the index to be changed is beyond bytearray length file set
            # it to the max index
            if (px_i > max_index):
                px_i = max_index

            byte_index = self.header_length + px_i
            new_bytes[byte_index] = int(amount * 256)

        self.new_bytes = new_bytes


    def save_image(self, name):
        """Save the image to a file. Keep trying by re-glitching the image with
        less iterations if it fails
        """

        while True:
            self.glitch_bytes()

            try:
                stream = io.BytesIO(self.new_bytes)
                im = Image.open(stream)
                im.save(name)
                return

            except (SyntaxError, IOError, AttributeError) as e:
                if self.iterations == 1:
                    raise JpegError(str(e))
                self.iterations -= 1


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['glitch_art_display']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=8.3.1,<9.0.0', 'click>=8.0.1,<9.0.0']

entry_points = \
{'console_scripts': ['glitch-art-display = glitch_art_display.main:main']}

setup_kwargs = {
    'name': 'glitch-art-display',
    'version': '1.0.1',
    'description': 'Generates .MP4 slideshow with seizure-inducing glitch transitions',
    'long_description': None,
    'author': 'TheTechromancer',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3

import os
import click
import hashlib
from .jpeg import *
import subprocess as sp
from pathlib import Path
from multiprocessing import cpu_count

from concurrent.futures import ThreadPoolExecutor

cache_dir = Path.home() / '.cache/glitch-art-display'
cache_dir.mkdir(exist_ok=True)


frame_timings = [5, 4, 3, 2, 2, 1, 1, 1, 1]


def is_cached(filename):

    if (cache_dir / filename).is_file():
        return True
    return False


def find_images(directory):

    valid_suffixes = ['.jpg', '.jpeg']
    convert_suffixes = ['.png']

    for root,dirs,files in os.walk(directory):
        for file in files:
            filename = Path(root) / file

            # if image is a JPEG
            if any([filename.suffix.lower().endswith(s) for s in valid_suffixes]):
                yield filename

            # if image is an unsupported image type
            elif any([filename.suffix.lower().endswith(s) for s in convert_suffixes]):
                # try to convert it with imagemagick
                try:
                    new_filename = cache_dir / (filename.stem + '.jpg')
                    sp.run(['convert', str(filename), str(new_filename)], check=True)
                    yield new_filename
                except (FileNotFoundError, sp.CalledProcessError) as e:
                    print(f'[!] Unsupported file: {filename.name}')
                    print(f'[!]  - please install imagemagick in order to use {filename.suffix} files')


def gen_frames(image_dir, output, glitch_amount=100, num_image_frames=25, num_transition_frames=30, shuffle=False):

    frame_groups = []
    frames_output = Path(output)
    frames_output.mkdir(exist_ok=True)
    frame_number = 0

    transition_frames = int(num_transition_frames/2)
    found_images = list(find_images(image_dir))
    if shuffle:
        random.shuffle(found_images)

    for image in found_images:

        print(f'[+] Generating frames for {image.name}')

        glitched_frames = []

        with ThreadPoolExecutor(max_workers=cpu_count()) as pool:

            image_bytes = bytearray(image.read_bytes())
            image_hash = hashlib.md5(image_bytes).hexdigest()
            png_filename = cache_dir / f'{image_hash}.png'
            pool.submit(sp.run, ['convert', str(image), str(png_filename)], check=True)

            # generate glitch frames
            glitched_futures = []
            for i in range(transition_frames):
                amount = int(glitch_amount * ((i+1)/transition_frames))
                glitched_futures.append(pool.submit(glitch, image, amount, i))

            pool.shutdown(wait=True)
            for glitched_future in glitched_futures:
                glitched_frame = glitched_future.result()
                if glitched_frame:
                    glitched_frames.append(glitched_frame)

        glitch_in = []
        random.shuffle(frame_timings)
        for i, glitched_image in enumerate(glitched_frames[::-1]):
            glitch_group = []
            for j in range(frame_timings[i % len(frame_timings)]):
                glitch_group.append(glitched_image)
            glitch_in.append(glitch_group)

        normal = []
        for i in range(num_image_frames):
            normal.append(png_filename)

        glitch_out = []
        random.shuffle(frame_timings)
        for i, glitched_image in enumerate(glitched_frames):
            glitch_group = []
            for j in range(frame_timings[i % len(frame_timings)]):
                glitch_group.append(glitched_image)
            glitch_out.append(glitch_group)

        frame_groups.append([glitch_in, normal, glitch_out])

    frames = []
    interlace_frames = max(2, min(int(transition_frames/4), 1))
    for i, [glitch_in, normal, glitch_out] in enumerate(frame_groups):
        for g in glitch_in:
            frames += g
        frames += normal
        if i == len(frame_groups)-1:
            for g in glitch_out:
                frames += g
        else:
            # truncate end of this glitch group
            glitch_out, glitch_interlaced = glitch_out[:-interlace_frames], glitch_out[-interlace_frames:]

            # truncate beginning of next glitch group
            frame_groups[i+1][0], glitch_next_interlaced = frame_groups[i+1][0][interlace_frames:], frame_groups[i+1][0][:interlace_frames]
            
            for g in glitch_out:
                frames += g
            for i in range(min(interlace_frames, len(glitch_interlaced))):
                frames += glitch_next_interlaced[i]
                frames += glitch_interlaced[i]

    for frame_number, frame in enumerate(frames):
        frame_name = frames_output / f'frame_{frame_number:09}.png'
        frame_name.unlink(missing_ok=True)
        frame_name.symlink_to(frame)

    print('[+] All frames generated.')


def glitch(image, amount=None, sequence=''):

    print(f'[+] Glitching {image.name} by {amount}')

    if amount is None:
        amount = random.randint(0,99)
    else:
        amount = max(0, min(99, int(amount)-1))

    image = Path(image)

    image_bytes = bytearray(image.read_bytes())
    image_hash = hashlib.md5(image_bytes).hexdigest()
    try:
        jpeg = Jpeg(image_bytes)
    except JpegError as e:
        print(f'[!] {e}')
        return

    jpeg_filename = cache_dir / f'{image_hash}_{amount}_{sequence}.jpg'
    png_filename = cache_dir / f'{image_hash}_{amount}_{sequence}.png'

    print(f'[+] Glitching {image.name} by {amount}')

    # checked for cached files
    if is_cached(png_filename):
        print(f'[+] Found cached frame for {png_filename}')

    else:
        while 1:
            try:
                jpeg.amount      = amount
                jpeg.seed        = random.randint(0,99)
                jpeg.iterations  = max(0, min(115, int(amount*1.15)))

                # create a new image if not cached
                jpeg.save_image(jpeg_filename)
                try:
                    sp.run(['convert', str(jpeg_filename), str(png_filename)], check=True)
                    break
                except Exception:
                    continue

            except JpegError as e:
                print(f'[!] {e}')
                continue

    return png_filename


def main():
    @click.command()
    @click.argument('input', required=True, type=click.Path(exists=True, file_okay=False, dir_okay=True))
    @click.argument('output', required=True, type=click.Path(exists=True, file_okay=False, dir_okay=True, writable=True))
    @click.option('--amount', default=50, type=click.IntRange(1, 100, clamp=True), help='Glitch amount (1 to 100)')
    @click.option('--normal-frames', default=25*25, type=int, help='Number of normal frames')
    @click.option('--transition-frames', default=30, type=int, help='Number of glitchy transition frames')
    @click.option('--shuffle', is_flag=True, help='Shuffle order of images')
    def go(input, output, amount, normal_frames, transition_frames, shuffle):
        gen_frames(input, output, amount, normal_frames, transition_frames, shuffle)
        click.echo(f"[+] Example: generate .MP4 with ffmpeg")
        click.echo(f"ffmpeg -framerate 25 -i {Path(output) / 'frame_%09d.png'} /tmp/output.mp4")

    go()

if __name__ == '__main__':
    main()